#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# 增/删/改/查/获取所有

# POST: {srv-name}/models
# DELETE: {srv-name}/models/{instID}
# POST: {srv-name}/models/{instID}
# GET: {srv-name}/models/{instID}
# GET: {srv-name}/models

import json
from pdb import set_trace as stop

from tornado.escape import json_decode
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application, RequestHandler


class CustomHandler(RequestHandler):

    # def get_argument(self, id: str):
    #     if self.request.headers['Content-Type'] == 'application/json':
    #         args = json_decode(self.request.body)
    #         return args[id]
    #     return super(CustomHandler, self).get_argument(id)

    def get_argument_dict(self):
        params = self.request.arguments
        for k, v in params.items():
            params[k] = v[0].decode("utf-8")
        return params

class Http():
    def __init__(self, config: dict, srv_name: str) -> None:
        self.config = config
        self.srv_name = srv_name
        self.srv = None

        # print('sqlite:///{0}/dm.db'.format(self.config["WORKDIR"]))
        # engine = create_engine('sqlite:///{0}/dm.db'.format(self.config["WORKDIR"]))
        # Entry.__table__.create(engine, checkfirst=True)
        # self.sql = sessionmaker(bind=engine)()

    def Build(self):
        self.appList = [
            # ('/{0}/models'.format(self.srv_name), ModelProc, dict(sql=self.sql)),
            # ('/{0}/models/(.*)'.format(self.srv_name),
            #  ModelProcID, dict(sql=self.sql)),
            # ('/{0}/models_exist/(.*)'.format(self.srv_name),
            #  ModelExist, dict(sql=self.sql)),
            # ('/{0}/property/(.*)'.format(self.srv_name),
            #  Property, dict(srv=self.srv)),
            
            ('/{0}/([a-zA-Z0-9_]+)'.format(self.srv_name),
             Action, dict(srv=self.srv)),
            
            # ('/{0}/event/(.*)'.format(self.srv_name), Event, dict(srv=self.srv)),
            # ('/{0}/mqtt/(.*)'.format(self.srv_name), MqttOp, dict(srv=self.srv)),
            # ('/{0}/devices'.format(self.srv_name),
            #  DevProc, dict(sql=self.sql, srv=self.srv)),
            # ('/{0}/devices/(.*)'.format(self.srv_name),
            #  DevProcID, dict(sql=self.sql, srv=self.srv)),
            # ('/{0}/file_upload'.format(self.srv_name),
            #  FileUploader, dict(srv=self.srv)),
            # ('/{0}/file_download'.format(self.srv_name),
            #  FileDownloader, dict(srv=self.srv)),
            # ('/{0}/file_remove'.format(self.srv_name),
            #  FileRemover, dict(srv=self.srv)),
            # ('/{0}/(.*)'.format(self.srv_name),
            #  ServiceProc, dict(srv=self.srv)),
        ]
        if self.config['PORT'] < 65535:
            self.httpApp = Application(self.appList)
            self.httpServer = HTTPServer(self.httpApp)
            self.httpServer.bind(self.config['PORT'])
            if 'HTTP_THREAD' in self.config:
                self.httpServer.start(self.config['HTTP_THREAD'])
            else:
                self.httpServer.start(1)
            
            
        # 0.1.20 MQTT support
        topic = '/{0}/+'.format(self.srv_name)
        self.mqtt = self.srv.mqtt
        self.mqtt.subscribe(topic)
        def callback (client, userdata, msg):
            try:
                topic = msg.topic
                payload = msg.payload.decode('utf-8')
                # check response
                # print("MQTT callback: {0}".format(topic))
                func_name = topic.split('/')[2]
                res = self.selector(func_name, payload)
                # Check responseTopic exists
                if hasattr(msg, 'properties') and hasattr(msg.properties, 'ResponseTopic'):
                    # print("MQTT respond: {0}".format(msg.properties.ResponseTopic))
                    self.mqtt.publish(msg.properties.ResponseTopic, json.dumps(res))
            except Exception as e:
                print(e)
        self.mqtt.message_callback_add(topic, callback)

    def Run(self):
        IOLoop.current().start()
            
    def selector(self, func_name: str, content=None):
        if (content == None or content == {} or content == ""):
            return getattr(self.srv.app, func_name)()
        else:
            if isinstance(content, str):
                return getattr(self.srv.app, func_name)(**(json.loads(content)))
            elif isinstance(content, dict):
                try:
                    return getattr(self.srv.app, func_name)(**content)
                except:
                    return getattr(self.srv.app, func_name)(content)
            else:
                # self.app.alarmer.alarm(Err.SRV_REQ_INVALID, 1,
                #                    "f:%{0} - {1}".format(func_name, content))
                return "Content type is not supported"

# class ServiceProc(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv

#     def get(self, func_name: str):
#         self.write(json.dumps(self.srv.Execute(func_name), ensure_ascii=False))
#         self.set_status(200)

#     def post(self, func_name: str):
#         content = self.get_argument('content')
#         self.write(json.dumps(self.srv.Execute(
#             func_name, content), ensure_ascii=False))
#         self.set_status(200)


# class FileUploader(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv

#     def post(self):
#         self.srv.app.upload(self.request)
#         self.set_status(200)


# class FileDownloader(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv

#     def post(self):
#         dataset = self.get_argument('dataset')
#         file_id = self.get_argument('file_id')
#         block = self.get_argument('block')
#         type = self.get_argument('type')
#         data = self.srv.app.download(file_id, dataset, block, type)
#         self.write(data)
#         self.set_status(200)


# class FileRemover(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv

#     def post(self):
#         dataset = self.get_argument('dataset')
#         file_id = self.get_argument('file_id')
#         block = self.get_argument('block')
#         type = self.get_argument('type')
#         self.srv.app.remove(file_id, dataset, block, type)
#         self.set_status(200)


# class ModelProc(CustomHandler):
#     def initialize(self, sql):
#         self.sql = sql

#     # Add new model
#     def post(self):
#         print(self.request.headers['Content-Type'])
#         model = self.sql.query(Entry).filter_by(
#             id=self.get_argument('id')).first()

#         if model == None:
#             # Add
#             model = Entry(id=self.get_argument('id'),
#                           content=self.get_argument('content'))
#             print("Entry Added")
#             self.sql.add(model)
#         else:
#             # Update
#             model.content = self.get_argument('content')
#             print("Entry exists, update")
#         self.sql.commit()
#         self.set_status(200)
#         self.write('{0}'.format(model))

#     # Get all models
#     def get(self):
#         print("Get Models")
#         model = self.sql.query(Entry).all()
#         self.set_status(200)
#         self.write("{0}".format(model))


# class ModelProcID(CustomHandler):
#     def initialize(self, sql):
#         self.sql = sql

#     # Delete model
#     def delete(self, id):
#         print("Delete Entry - ID: " + id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         self.sql.delete(model)
#         self.sql.commit()
#         self.set_status(200)
#         self.write("{0}".format(model))

#     # Update model
#     def post(self, id):
#         print("Update Entry - ID: " + id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         model.content = self.get_argument('content')
#         self.sql.commit()
#         self.set_status(200)
#         self.write("{0}".format(model))

#     # Get model
#     def get(self, id: str):
#         print("Get Entry - ID: " + id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         self.set_status(200)
#         self.write("{0}".format(model))


# class ModelExist(CustomHandler):
#     def initialize(self, sql):
#         self.sql = sql

#     # Check exist
#     def get(self, id: str):
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         self.set_status(200)
#         if model == None:
#             print("Entry Not Exist - ID: " + id)
#             self.write("false")
#         else:
#             print("Entry Exist - ID: " + id)
#             self.write("true")


# 读取/操作

# GET: {srv-name}/property/{property-name}
# POST: {srv-name}/action/{function-name}
# POST & MQTT: {srv-name}/event/{event-name}

# class Property(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv

#     def post(self, key: str):
#         content = self.get_argument('content')
#         is_read = self.get_argument('is_read')
#         self.set_header('Access-Control-Allow-Origin', '*')
#         if is_read:
#             self.write(json.dumps(self.srv.ReadProperty(
#                 key, content), ensure_ascii=False))
#         else:
#             self.write(json.dumps(self.srv.WriteProperty(
#                 key, content), ensure_ascii=False))


class Action(CustomHandler):
    def initialize(self, srv):
        self.srv = srv

    def get(self, func_name: str):
        self.write(json.dumps(self.srv.Execute(func_name), ensure_ascii=False))

    def post(self, func_name: str):
        args = self.get_argument_dict()
        self.set_header('Access-Control-Allow-Origin', '*')
        self.write(json.dumps(self.selector(func_name, args)))

    def selector(self, func_name: str, content=None):
        if (content == None or content == {} or content == ""):
            return getattr(self.srv.app, func_name)()
        else:
            if isinstance(content, str):
                return getattr(self.srv.app, func_name)(**(json.loads(content)))
            elif isinstance(content, dict):
                try:
                    return getattr(self.srv.app, func_name)(**content)
                except:
                    return getattr(self.srv.app, func_name)(content)
            else:
                # self.app.alarmer.alarm(Err.SRV_REQ_INVALID, 1,
                #                    "f:%{0} - {1}".format(func_name, content))
                return "Content type is not supported"

# class Event(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv
#         self.mqtt = srv.mqtt

#     def post(self, event_name: str):
#         setup = int(self.get_argument('setup'))  # 0 = off, 1 = on
#         pub_topic = self.get_argument('pub_topic')
#         qos = 1
#         try:
#             qos = int(self.get_argument('qos'))
#         except:
#             pass

#         print("Event Subscribed: {0}, {1}".format(setup, pub_topic))

#         def on_event_callback(content):
#             if (setup > 0):
#                 print("Event auto-pubed: {0}, {1}".format(setup, pub_topic))
#                 self.mqtt.publish(pub_topic, json.dumps(
#                     {'content': content}, ensure_ascii=False), qos)

#         if not hasattr(self.srv.app, 'ecbs'):
#             print("ecb not initiated")
#             self.write("Error: ecb not initiated")
#         else:
#             self.srv.app.ecbs[event_name] = on_event_callback
#             self.write("OK")

# MQTT 操作 (Flow)


# class MqttOp(CustomHandler):
#     def initialize(self, srv):
#         self.srv = srv
#         self.mqtt = srv.mqtt

#     def post(self, func_name: str):
#         op = self.get_argument('op')
#         sub_topic = None

#         try:
#             sub_topic = self.get_argument('sub_topic')
#         except:
#             pass

#         pub_topic = self.get_argument('pub_topic')
#         qos = 1

#         try:
#             qos = int(self.get_argument('qos'))
#         except:
#             pass

#         print("MQTT Op: {0}, {1}, {2}".format(op, sub_topic, pub_topic))

#         # Automate: sub func input, pub func output
#         if (op == "add"):
#             if (sub_topic != None):
#                 self.mqtt.subscribe(sub_topic, qos)

#             # Inject constants
#             constants = {}
#             try:
#                 constants = json.loads(self.get_argument('constants'))
#             except:
#                 pass

#             def on_subs_callback(client, userdata, msg):
#                 msg.payload = json.loads(msg.payload)
#                 msg.payload.update(constants)

#                 result = self.srv.Execute(func_name, msg.payload)
#                 if result is not None:
#                     # Decode input
#                     self.mqtt.publish(pub_topic, json.dumps(
#                         {'content': self.srv.Execute(func_name, msg.payload)}, ensure_ascii=False), qos)
#             # Sub input
#             self.mqtt.message_callback_add(sub_topic, on_subs_callback)
#         elif (op == "remove"):
#             self.mqtt.unsubscribe(sub_topic)
#             self.mqtt.message_callback_remove(sub_topic)
#             # 0.1.19
#             if sub_topic in self.srv.gcbs:
#                 print("Remove global callback array at " +
#                       sub_topic)
#                 del self.srv.gcbs[sub_topic]
#         self.write("OK")

# # Protocol 协议服务专用


# class DevProc(CustomHandler):
#     def initialize(self, sql, srv):
#         self.sql = sql
#         self.srv = srv

#     # Add new model
#     def post(self):
#         print(self.request.headers['Content-Type'])

#         req = json.loads(self.request.body)
#         req_json = json.dumps(req, ensure_ascii=False)
#         id = req['id']

#         model = self.sql.query(Entry).filter_by(id=id).first()

#         if model == None:
#             # Add
#             model = Entry(id=id, content=req_json)
#             print("Device Added")
#             res = self.srv.AddDev(req)
#             # if res is not None:
#             #     self.write(str(res))
#             #     return
#             self.sql.add(model)
#         else:
#             # Update
#             res = self.srv.UpdateDev(req)
#             # if res is not None:
#             #     self.write(str(res))
#             #     return
#             model.content = req_json
#             print("Device exists, update")
#         self.sql.commit()
#         self.set_status(200)
#         self.write('Device Added: {0}'.format(id))

#     # Get all models
#     def get(self):
#         models = self.sql.query(Entry).all()
#         for m in models:
#             m.content = json.dumps(json.loads(m.content), ensure_ascii=False)
#         models = json.dumps([{"id": m.id, "content": m.content}
#                             for m in models], ensure_ascii=False)
#         self.set_status(200)
#         self.write(models)


# class DevProcID(CustomHandler):
#     def initialize(self, sql, srv):
#         self.sql = sql
#         self.srv = srv

#     # Delete model
#     def delete(self, id):
#         print("Delete Device - ID: " + id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         if model is None:
#             self.write('Device does not exist - ID: ' + id)

#         res = self.srv.DeleteDev(id)
#         if res is not None:
#             print(res)
#             self.write(str(res))
#             return

#         self.sql.delete(model)
#         self.sql.commit()
#         self.set_status(200)
#         self.write('Device Deleted: {0}'.format(id))
#         if res == "exit":
#             self.set_status(200)
#             self.write('REBOOT, Device Deleted: {0}'.format(id))
#             import os
#             os._exit(1)

#     # Get model
#     def get(self, id: str):
#         print("Get Device - ID: " + id)
#         m = self.sql.query(Entry).filter_by(id=id).first()
#         model = json.dumps(
#             {"id": m.id, "content": m.content}, ensure_ascii=False)
#         self.set_status(200)
#         self.write(model)



import logging, sys
class CustomFormatter(logging.Formatter):
    grey = "\033[33;1;30m"
    cyan = "\033[33;1;36m"
    yellow = "\033[33;1;33m"
    red = "\033[33;1;31m"
    bold_red = "\033[33;7;31m"
    reset = "\033[0m"
    format = '%(asctime)s: %(message)s'

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: cyan + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt='%Y-%m-%d %H:%M:%S')
        return formatter.format(record)
    
logger = logging.getLogger("mtnode")
logger.setLevel(logging.DEBUG)
stdout = logging.StreamHandler(stream=sys.stdout)
stdout.setFormatter(CustomFormatter())
logger.addHandler(stdout)

if __name__ == '__main__':
    logger.info("This is a test")
    logger.debug("This is a test")
    logger.warning("This is a test")
    logger.error("This is a test")
    logger.critical("This is a test")


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from metathing.service import Service
from metathing.logger import logger
import socket, time
import struct

MCAST_GRP = '239.1.2.3'
MCAST_PORT = 34567
RECV_MCAST_PORT = 34568

class MetaThing():
    def __init__(self, config: object, srv_name: str):
        self.srv = Service(config, srv_name)
        self.addr = None

    def Bind(self, app:object):
        self.srv.mqtt.connect()
        self.srv.Bind(app)
        app.LoadModel()

    def Run(self):
        self.srv.http.Run()
        
    def OpenIPBroadcast(self):
        import threading
        t = threading.Thread(target=MetaThing.MTNodeIPSend)
        logger.info('[Auth]: Starting IP broadcast thread...')
        t.start()

    @staticmethod
    def KillProcessWithSameIpAndPort(ip, port):
        import psutil
        current_pid = psutil.Process().pid 
        for conn in psutil.net_connections():
            if conn.laddr.ip == ip and conn.laddr.port == port and conn.pid != current_pid:
                pid = conn.pid
                psutil.Process(pid).kill()
                logger.warning(f"Killed conflicted process with pid {pid}")
        
    @staticmethod
    def MTNodeIPRequest():
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        req_msg = 'MTNode_IP_REQUEST'

        sock_recv = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock_recv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock_recv.bind(('', RECV_MCAST_PORT))
        mreq = struct.pack("4sl", socket.inet_aton(MCAST_GRP), socket.INADDR_ANY)
        sock_recv.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        sock_recv.settimeout(5)
        while True:
            sock.sendto(req_msg.encode(), (MCAST_GRP, MCAST_PORT))
            try:
                logger.info('Requesting MTNode IP...')
                data, address = sock_recv.recvfrom(1024)
                data = data.decode()
                logger.info(f'Success: {data}')
                sock.close()
                sock_recv.close()
                return data
            except socket.timeout:
                logger.info('Timeout, retrying...')
        
    @staticmethod        
    def MTNodeIPSend():
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        ready = False;
        
        while not ready:  
            # attempt to reconnect, otherwise sleep for 2 seconds  
            try:  
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind(('', MCAST_PORT))
                mreq = struct.pack("4sl", socket.inet_aton(MCAST_GRP), socket.INADDR_ANY)
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
                sock_sender = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
                sock_sender.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
                csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                csock.connect(('8.8.8.8', 80))
                (addr, port) = csock.getsockname()
                csock.close()
                ready = True
            except:  
                time.sleep(2)  
                
        while True:
            data, address = sock.recvfrom(1024)
            if data == b'MTNode_IP_REQUEST':
                logger.info(f'Received request from: {address}')
                logger.info(f'Sending my address: {addr}')
                sock_sender.sendto(addr.encode(), (MCAST_GRP, RECV_MCAST_PORT))
                
# if __name__ == '__main__':
#     # Open a thread to run MetaThing.MTNodeIPSend()
#     import threading
#     t = threading.Thread(target=MetaThing.MTNodeIPSend)
#     t.start()
    
#     MetaThing.MTNodeIPRequest()     


import random
from pdb import set_trace as stop

from paho.mqtt import client as mqtt_client
from paho.mqtt.properties import Properties, PacketTypes
import threading, time
from metathing.logger import logger

class Mqtt():
    def __init__(self, broker: str = '127.0.0.1', port: int = 1883, ws_addr: str = None, ws_port = 8083) -> None:
        self.broker = broker
        self.port = port
        self.client = None
        self.client_2 = None
        self.resp_topic = "/resp/{0}".format(random.randint(0, 999999999999))
        self.ws_addr = ws_addr
        self.ws_port = ws_port

    def connect(self):
        def on_connect(client, userdata, flags, rc, properties=None):
            if rc == 0:
                logger.info("Connected: {0}".format(client._client_id.decode()))
                pass
            else:
                logger.error("Failed to connect, return code %d\n" % rc)
        if self.ws_addr is not None:
            self.client = mqtt_client.Client(
                f'ws-srv-{random.randint(0, 99999999999)}', protocol=5, transport="websockets")
            self.client_2 = mqtt_client.Client(
                f'ws-srv2-{random.randint(0, 99999999999)}', protocol=5, transport="websockets")
            logger.info("WebSocket ID: {0}".format(self.client._client_id.decode()))
            self.client.on_connect = on_connect
            self.client.ws_set_options(path='/mqtt')
            self.client_2.ws_set_options(path='/mqtt')
            try:
                self.client.connect(self.ws_addr, port=self.ws_port)
                self.client_2.connect(self.ws_addr, port=self.ws_port)
            except:
                time.sleep(1)
                self.client.connect(self.ws_addr, port=self.ws_port)
                self.client_2.connect(self.ws_addr, port=self.ws_port)
        else:
            self.client = mqtt_client.Client(
                f'mqtt-srv-{random.randint(0, 99999999999)}', protocol=5)
            self.client_2 = mqtt_client.Client(
                f'mqtt-srv2-{random.randint(0, 99999999999)}', protocol=5)
            logger.info("MQTT ID: {0}".format(self.client._client_id.decode()))
            self.client.on_connect = on_connect
            try:
                self.client.connect(self.broker, self.port)
                self.client_2.connect(self.broker, self.port)
            except:
                time.sleep(1)
                self.client.connect(self.broker, self.port)
                self.client_2.connect(self.broker, self.port)
        self.client.subscribe(self.resp_topic, 1, properties=Properties(PacketTypes.SUBSCRIBE))
        self.client.loop_start()
        
        self.client_2.subscribe(self.resp_topic, 1, properties=Properties(PacketTypes.SUBSCRIBE))
        self.client_2.loop_start()

    def publish(self, topic: str, content: str, qos: int = 0, retain: bool = False, properties=Properties(PacketTypes.PUBLISH)):
        self.client_2.publish(topic, content, qos, retain, properties=properties)

    # Sync
    # def publish_request(self, topic: str, content: str, qos: int = 0, timeout = 5, retain: bool = False):
    #     resp_lock = threading.Lock()
    #     resp_lock.acquire()
    #     properties = Properties(PacketTypes.PUBLISH)
    #     properties.ResponseTopic = self.resp_topic
    #     self.resp = None
        
    #     def on_response(client, userdata, msg):
    #         self.resp = msg.payload.decode('utf-8')
    #         self.client.message_callback_remove(self.resp_topic)
    #         if resp_lock.locked():
    #             resp_lock.release()
        
    #     self.client.message_callback_add(self.resp_topic, on_response)
    #     # Send the request
    #     self.client.publish(topic, content, qos, retain, properties=properties)
    #     resp_lock.acquire(blocking=True, timeout=timeout)
    #     if resp_lock.locked():
    #         resp_lock.release()
    #     return self.resp
        

    def publish_request(self, topic: str, content: str, qos: int = 0, timeout = 5, retain: bool = False):
        try:
            resp_lock = threading.Lock()
            resp_lock.acquire()
            properties = Properties(PacketTypes.PUBLISH)
            properties.ResponseTopic = self.resp_topic
            self.resp = None
            
            def on_response(client, userdata, msg):
                self.resp = msg.payload.decode('utf-8')
                self.client_2.message_callback_remove(self.resp_topic)
                if resp_lock.locked():
                    resp_lock.release()
                else:
                    logger.warning("Message reached but lock is timeout! （考虑缩短下游耗时或提高超时时间？）")
            
            self.client_2.message_callback_add(self.resp_topic, on_response)
            # Send the request
            self.client_2.publish(topic, content, qos, retain, properties=properties)
            resp_lock.acquire(blocking=True, timeout=timeout)
            if resp_lock.locked():
                resp_lock.release()
            return self.resp
        except Exception as e:
            logger.Error(e)
        
    # NEED TEST
    def publish_request_async(self, res, topic: str, content: str, qos: int = 0, timeout = 5, retain: bool = False):
        properties = Properties(PacketTypes.PUBLISH)
        properties.ResponseTopic = self.resp_topic
        self.resp = None
        def on_response(client, userdata, msg):
            self.resp = msg.payload.decode('utf-8')
            self.client_2.message_callback_remove(self.resp_topic)
            res = self.resp
        self.client_2.message_callback_add(self.resp_topic, on_response)
        self.client_2.publish(topic, content, qos, retain, properties=properties)
        

    # Accept list, e.g. subscribe([("my/topic", 0), ("another/topic", 2)])
    def subscribe(self, topic, qos: int = 1):
        if (isinstance(topic, list)):
            self.client.subscribe(topic, qos, properties=Properties(PacketTypes.SUBSCRIBE))
            # for t in topic:
            #     print("Subscribe to %s" % t)
        else:
            # print("Subscribe to %s" % topic)
            self.client.subscribe(topic, qos, properties=Properties(PacketTypes.SUBSCRIBE))

    def unsubscribe(self, topic):
        # print("Unsubscribe %s" % topic)
        self.client.unsubscribe(topic)

    def message_callback_add(self, topic: str, callback):
        # print("Add callback to topic %s" % topic)
        self.client.message_callback_add(topic, callback)

    def message_callback_remove(self, topic: str):
        # print("Remove callback to topic %s" % topic)
        self.client.message_callback_remove(topic)

    def stop(self):
        self.client.loop_stop()
        self.client.disconnect()
        logger.info("Disconnected from MQTT Broker: %s:%s!" %
              (self.broker, self.port))



# #! /usr/bin/env python3
# # -*- coding: utf-8 -*-
# from paho.mqtt import client as mqtt_client
# import random
# import time
# from pdb import set_trace as stop
# import json

# #【弃用 Alarmer】

# class Err:
#     SRV_NO_EXIST = 900 # 无此服务
#     SRV_NO_RESP = 901 # 无回应
#     SRV_REQ_INVALID = 902 # 无效 Req
#     SRV_RESP_INVALID = 903 # 无效 Res

#     DEV_NO_CONN = 1000 # 无法连接设备
#     DEV_NO_RESP = 1001 # 无法得到回应
#     DEV_REQ_INVALID = 1002 # 无效的 Request
#     DEV_RESP_INVALID = 1003 # 无效的 Response

#     BLE_RESP_INVALID = 11003 # 无效蓝牙回应
#     BLE_NO_CHAR = 11100 # 无效蓝牙特征

# class Alarmer():
#     def __init__(self, broker: str = 'broker.emqx.io', port: int = 1883) -> None:
#         self.broker = broker
#         self.port = port
#         self.client = None

#     def connect(self, srv_id, _dev_id, label, srv_type):
#         def on_connect(client, userdata, flags, rc):
#             if rc == 0:
#                 print("[Alarmer] Connected to MQTT Broker: %s:%s!" % (self.broker, self.port))
#             else:
#                 print("[Alarmer] Failed to connect, return code %d\n" % rc)

#         self.client = mqtt_client.Client(f'python-alarmer-{random.randint(0, 99999999999)}')
#         self.client.on_connect = on_connect
#         self.client.connect(self.broker, self.port)
#         self.client.loop_start()
#         self.topic = 'alarm/{0}/{1}'.format(srv_id, _dev_id)
#         self.alarm_msg = {
#             "srv_id": srv_id,
#             "_dev_id": _dev_id,
#             "label": label,
#             "srv_type": srv_type,
#             "code_err": -1,
#             "code_severity": -1,
#             "time": 0,
#             "content": ""
#         }

#     def alarm(self, code_err: int, code_severity: int, content: str):
#         self.alarm_msg['code_err'] = code_err
#         self.alarm_msg['code_severity'] = code_severity
#         self.alarm_msg['content'] = content
#         self.alarm_msg['time'] = time.time_ns()
#         msg = json.dumps(self.alarm_msg, ensure_ascii=False)
#         print("[ALARM]: " + msg)
#         self.client.publish(self.topic, msg, 1, False)


import json
from metathing.logger import logger
import inspect
from functools import partial
from pdb import set_trace as stop
import sys

class Application():
    def __init__(self):
        self.srv = None
        self.Dev = None
        self.event_sender = {}
        self.devs = {}

    def AddDev(self, model):
        if self.Dev is None:
            raise Exception("Device model not defined!")
        dev_id = model['dev_id']
        if 'init_config' in model:
            config = model['init_config']
        config['dev_id'] = dev_id
        self.devs[dev_id] = self.Dev(self.srv, config)
        
    def UpdateDev(self, model):
        self.AddDev(model)

    def DeleteDev(self, id):
        if (id in self.devs):
            print("Releasing: "+id)
            self.devs[id].release()
            del self.devs[id]

    def DevsInfo(self):
        devs_info = {}
        for _dev_id, dev in self.devs.items():
            devs_info[_dev_id] = dev.model
        return devs_info
        
    def Event(func):
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            # Get function name
            fn_name = func.__name__
            if fn_name in self.event_sender:
                for topic in self.event_sender[fn_name]:
                    self.srv.mqtt.publish(topic, json.dumps(result))
            return result
        return wrapper
    
        
    def BindAction(self, model):
        fn_name = model['tag']
        # Check action
        if hasattr(self, fn_name) == False:
            logger.error("Action not found [{0}]".format(fn_name))
            return False
        fn = getattr(self, fn_name)
        
        # Process overrides
        vo_map = {}
        if 'var_override' in model:
            for vo in model['var_override']:
                vo_map[vo['var_uid']] = vo['default']
                
        input_args = list(inspect.signature(fn).parameters.keys())
        input_args_const = {}
        # Check inputs
        for arg in model['srv_input']:
            if not arg['var_tag'] in input_args:
                logger.error("Input not found [{0}]".format(arg['var_tag']))
                return False
            arg['is_constant'] = arg['uid'] in vo_map
            if arg['is_constant']:
                v = self._dtype_cast(arg['dtype'], vo_map[arg['uid']])  
                input_args_const[arg['var_tag']] = v
            else: 
                self._dtype_cast(arg['dtype'], arg['default'])
        
        # Generate lambda that replace input with default value if is_constant
        if len(input_args_const) > 0:
            new_fn = partial(fn, **input_args_const)
        else:
            new_fn = fn
        
        # Subscribe to new topic
        self.srv.mqtt.subscribe(model['topic'])
        def callback (client, userdata, msg):
            try:
                message = []
                payload = msg.payload.decode('utf-8')
                inputs = json.loads(payload)
                res = new_fn(**inputs)
                message.append(f"-= {msg.topic} =-")
                message.append(f"- In -")
                for k, v in inputs.items():
                    message.append(f"{k}: {v}")
                message.append(f"- Out -")
                for k,v in res.items():
                    message.append(f"{k}: {v}")
                if hasattr(msg, 'properties') and hasattr(msg.properties, 'ResponseTopic'):
                    message.append(f"- ResponseTopic -")
                    message.append(f"{msg.properties.ResponseTopic}")
                    self.srv.mqtt.publish(msg.properties.ResponseTopic, json.dumps(res))
                self._print_frame(message)
            except Exception as e:
                print(e)
        self.srv.mqtt.message_callback_add(model['topic'], callback)
        return True
    
    def BindEvent(self, model):
        fn_name = model['tag']
        # Check action
        if hasattr(self, fn_name) == False:
            logger.error("Event not found [{0}]".format(fn_name))
            return False
        
        if fn_name in self.event_sender:
            self.event_sender[fn_name].append(model['topic'])
        else:
            self.event_sender[fn_name] = [model['topic']]
        return True
    
    def LoadModel(self, gdb_topic = "/auth/query_adapter_config"):
        try:
            res = self.srv.mqtt.publish_request(gdb_topic, json.dumps({'data': {'srv_id': self.srv.srv_name}}))
            self._processing_model(res)
            return True
        except Exception as e:
            logger.error("Load model error: {0}".format(e))
            print(e)
            return None
              
    @staticmethod
    def _dtype_cast(dtype: str, item: str):
        if dtype == 'int':
            return int(item)
        elif dtype == 'float':
            return float(item)
        elif dtype == 'bool':
            if item == 'true':
                return True
            else:
                return False
        return item
    
    def _processing_model(self, res: str):
        try:
            if res == 'null' or res is None:
                self._print_frame(["* Service has not been registered in MTNode *"])
                return None
            model = json.loads(res)
            self.model = model
            message = [
                f"-= {self.srv.srv_name} =-",
                f"UID: {model['uid']}",
            ]
            
            if 'init_config' in model:
                message.append("")
                message.append(f"- Config -")
                for k, v in model['init_config'].items():
                    message.append(f"{k}: {v}")
                    setattr(self, k, v)
                    
            if 'device_model' in model and len(model['device_model']) > 0:
                for device in model['device_model']:
                    self.AddDev(device)
                    message.append("")
                    message.append(f"Device: [{device['dev_id']}]")
                    if 'init_config' in device:
                        for k, v in device['init_config'].items():
                            message.append(f"{k}: {v}")
            
            message.append('')
            message.append('- Registered Actions and Events -')
            
            name_map = {
                'srv_act': 'Global Actions',
                'srv_evt': 'Global Events',
                'dev_act': 'Device Actions',
                'dev_evt': 'Device Events',
            }
            for key in ['srv_act', 'srv_evt', 'dev_act', 'dev_evt']:                
                if key in model and len(model[key]) > 0:
                    message.append("")
                    message.append(f"{name_map[key]}: [{len(model[key])}] items")
                    for item in model[key]:
                        if key == 'srv_act': self.BindAction(item) 
                        if key == 'srv_evt': self.BindEvent(item) #TODO: Need test
                        if key == 'dev_act': 
                            if not 'dev_id' in item:
                                logger.error("Device ID miss-linked （设备ID绑定错误，可能是因为适配服务被改动而实例服务未更新）")
                            dev_id = item['dev_id']
                            if not dev_id in self.devs:
                                logger.error("Device not found [{0}]".format(dev_id))
                            self.devs[dev_id].BindAction(item)
                        if key == 'dev_evt':
                            if not 'dev_id' in item:
                                logger.error("Device ID miss-linked （设备ID绑定错误，可能是因为适配服务被改动而实例服务未更新）")
                            dev_id = item['dev_id']
                            if not dev_id in self.devs:
                                logger.error("Device not found [{0}]".format(dev_id))
                            self.devs[dev_id].BindEvent(item)
                        message.append(f"{item['tag']}({item['topic']}): In[{len(item['srv_input']) if 'srv_input' in item and item['srv_input'] else 0 }] Out[{len(item['srv_output']) if 'srv_output' in item and item['srv_output'] else 0 }]")
            self._print_frame(message)
        except Exception as e:
            logger.error("Load model error: {0}".format(e))
            print(e)
            return None
              
    @staticmethod
    def _print_frame(content):
        frame_width = len(max(content, key=len)) + 4
        frame = '+' + '-' * (frame_width) + '+'
        logger.info(frame)
        for line in content:
            spaces_before = int((frame_width - len(line)) / 2)
            spaces_after = frame_width - len(line) - spaces_before
            logger.info('|' + ' ' * spaces_before + line + ' ' * spaces_after + '|')
        logger.info(frame)
              
    # Customized functions
    def ping(self):
        return "OK"
          
    def release(self):
        return


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# 增/删/改/查/获取所有

# POST: {srv-name}/models
# DELETE: {srv-name}/models/{instID}
# POST: {srv-name}/models/{instID}
# GET: {srv-name}/models/{instID}
# GET: {srv-name}/models

import json
from pdb import set_trace as stop

from tornado.escape import json_decode
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.routing import (AnyMatches, DefaultHostMatches, HostMatches,
                             ReversibleRouter, ReversibleRuleRouter, Rule,
                             URLSpec, _RuleList)
from tornado.web import Application, RequestHandler, _ApplicationRouter
from .mqtt import Mqtt
from functools import partial
import inspect
import sys

from metathing.logger import logger

# from .alarmer import Alarmer, Err

class CustomHandler(RequestHandler):
    def get_argument(self, id: str):
        if self.request.headers['Content-Type'] == 'application/json':
            args = json_decode(self.request.body)
            return args[id]
        return super(CustomHandler, self).get_argument(id)

    def get_argument_dict(self):
        params = self.request.arguments
        for k, v in params.items():
            params[k] = v[0].decode("utf-8")
        return params

# Base for protocol service


class Device():
    def __init__(self, srv, model) -> None:
        self.config = srv.config
        self.srv_name = srv.srv_name
        # self.resources = {}
        self.event_sender = {}
        if isinstance(model, str):
            self.model = json.loads(model)
        else:
            self.model = model

        self.srv = srv
        self.mqtt = Mqtt(srv.config['MQTT_ADDR'], srv.config['MQTT_PORT'])
        self.mqtt.connect()
        # self.cb_stack = {}
        
        for k, v in self.model.items():
            setattr(self, k, v)
            
        # self.alarmer = None
        self._dev_id = model['dev_id']
        self.build()

            # if 'srv_type' in model:
            #     self.srv_type = model['srv_type']
            # else:
            #     self.srv_type = "Device"

            # if 'label' in model:
            #     self.label = model['label']
            # else:
            #     self.label = "N/A"

            # self.alarmer = Alarmer(self.mqtt.broker, self.mqtt.port)
            # self.alarmer.connect(self.srv_name, self._dev_id,
            #                      self.label, self.srv_type)

        # self.dev_srvs = {}
        # if 'services' in model:
        #     for s in model['services']:
        #         self.dev_srvs[s['id']] = s

    def initialize(self):
        raise NotImplementedError

    #region Built-in funcs
    def DevInfo(self):
        return {'info': self.model}
    
    def Event(func):
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            # Get function name
            fn_name = func.__name__
            if fn_name in self.event_sender:
                for topic in self.event_sender[fn_name]:
                    self.srv.mqtt.publish(topic, json.dumps(result))
            return result
        return wrapper
    
        
    def BindAction(self, model):
        fn_name = model['tag']
        # Check action
        if hasattr(self, fn_name) == False:
            logger.error("Device Action not found [{0}]".format(fn_name))
            return False
        fn = getattr(self, fn_name)
        
        # Process overrides
        vo_map = {}
        if 'var_override' in model:
            for vo in model['var_override']:
                vo_map[vo['var_uid']] = vo['default']
                
        input_args = list(inspect.signature(fn).parameters.keys())
        input_args_const = {}
        # Check inputs
        for arg in model['srv_input']:
            if not arg['var_tag'] in input_args:
                logger.error("Input not found [{0}]".format(arg['var_tag']))
                return False
            arg['is_constant'] = arg['uid'] in vo_map
            if arg['is_constant']:
                v = self._dtype_cast(arg['dtype'], vo_map[arg['uid']])  
                input_args_const[arg['var_tag']] = v
            else: 
                self._dtype_cast(arg['dtype'], arg['default'])
        
        # Generate lambda that replace input with default value if is_constant
        if len(input_args_const) > 0:
            new_fn = partial(fn, **input_args_const)
        else:
            new_fn = fn
        
        # Subscribe to new topic
        self.srv.mqtt.subscribe(model['topic'])
        def callback (client, userdata, msg):
            try:
                message = []
                payload = msg.payload.decode('utf-8')
                inputs = json.loads(payload)
                res = new_fn(**inputs)
                message.append(f"-= {msg.topic} =-")
                message.append(f"- In -")
                for k, v in inputs.items():
                    message.append(f"{k}: {v}")
                message.append(f"- Out -")
                if res is not None:
                    for k,v in res.items():
                        message.append(f"{k}: {v}")
                    if hasattr(msg, 'properties') and hasattr(msg.properties, 'ResponseTopic'):
                        message.append(f"- ResponseTopic -")
                        message.append(f"{msg.properties.ResponseTopic}")
                        self.srv.mqtt.publish(msg.properties.ResponseTopic, json.dumps(res))
                self.srv.app._print_frame(message)
            except Exception as e:
                logger.error(e)
        self.srv.mqtt.message_callback_add(model['topic'], callback)
        return True
    
    def BindEvent(self, model):
        fn_name = model['tag']
        # Check action
        if hasattr(self, fn_name) == False:
            logger.error("Event not found [{0}]".format(fn_name))
            return False
        
        if fn_name in self.event_sender:
            self.event_sender[fn_name].append(model['topic'])
        else:
            self.event_sender[fn_name] = [model['topic']]
        return True
    
    @staticmethod
    def _dtype_cast(dtype: str, item: str):
        if dtype == 'int':
            return int(item)
        elif dtype == 'float':
            return float(item)
        elif dtype == 'bool':
            if item == 'true':
                return True
            else:
                return False
        return item
    
              
    #endregion

    def build(self):
        self.appList = [
            ('/{0}/devices/{1}/ping'.format(self.srv_name, self._dev_id), Ping),
            # ('/{0}/devices/{1}/resources/(.*)'.format(self.srv_name,
            #  self._dev_id), DevProperty, dict(dev=self, srv=self.srv)),
            # ('/{0}/devices/{1}/services/(.*)'.format(self.srv_name,
            #  self._dev_id), DevService, dict(dev=self, srv=self.srv)),
            # ('/{0}/devices/{1}/events/(.*)'.format(self.srv_name,
            #  self._dev_id), DevEvent, dict(dev=self, srv=self.srv)),
            ('/{0}/devices/{1}/selectors/(.*)'.format(self.srv_name,
             self._dev_id), DevSelector, dict(dev=self, srv=self.srv)),
            # ('/{0}/devices/{1}/mqtt/(.*)'.format(self.srv_name,
            #  self._dev_id), DevMqttOp, dict(dev=self, srv=self.srv)),
        ]
        
        if self.config['PORT'] < 65535:
            # 0.1.19 Fix release handlers
    #        self.srv.http.httpApp.add_handlers(r".*", self.appList)
            host_matcher = HostMatches(r".*")
            self.rule = Rule(host_matcher, _ApplicationRouter(
                self.srv.http.httpApp, self.appList))
            self.srv.http.httpApp.default_router.rules.insert(-1, self.rule)
            
        # 0.1.20 MQTT support
        topic = '/{0}/devices/{1}/selectors/#'.format(self.srv_name, self._dev_id)
        self.mqtt.subscribe(topic)
        def callback (client, userdata, msg):
            topic = msg.topic
            payload = msg.payload.decode('utf-8')
            # check response
            func_name = topic.split('/')[5]
            res = self.selector(func_name, payload)
            message = [
                "Message received: {0} - {1}".format(topic, payload),
                "Response: {0}".format(res)
            ]
            
            # Check responseTopic exists
            if hasattr(msg, 'properties') and hasattr(msg.properties, 'ResponseTopic'):
                self.mqtt.publish(msg.properties.ResponseTopic, json.dumps(res))
        self.mqtt.message_callback_add(topic, callback)

    def release(self):
        self.srv.http.httpApp.default_router.rules.remove(self.rule)

    # Selector
    def selector(self, func_name: str, content=None):
        try:
            if (content == None or content == '{}'):
                return getattr(self, func_name)()
            else:
                if isinstance(content, str):
                    return getattr(self, func_name)(**(json.loads(content)))
                elif isinstance(content, dict):
                    try:
                        return getattr(self, func_name)(**content)
                    except:
                        return getattr(self, func_name)(content)
                else:
                    # self.alarmer.alarm(Err.SRV_REQ_INVALID, 1,
                                    # "f:%{0} - {1}".format(func_name, content))
                    return "Content type is not supported"
        except Exception as e:
            return "Selector error: %s" % e

    # def service(self, srv_name: str, content=None):
    #     if self.dev_srvs is None:
    #         return "Device has no service"
    #     if srv_name in self.dev_srvs:
    #         srv_model = self.dev_srvs[srv_name]
    #         func_name = srv_model['selector']
    #         return self.selector(func_name, content)
    #     else:
    #         # self.alarmer.alarm(Err.SRV_NO_EXIST, 1, srv_name)
    #         return "Device does not have service: %s" % srv_name

    def ping(self):
        return "OK"
    
    # 【弃用管道】
    # def flow(self, op: str, srv_name='', sub_topic: str = None, pub_topic: str = None, qos=1):
    #     qos = int(qos)
    #     print("Add flow, dev addr: {0}".format(self))
    #     # Automate: sub func input, pub func output
    #     if (op == "add"):
    #         if (sub_topic != None):
    #             self.mqtt.subscribe(sub_topic, qos)

    #         # Inject constants
    #         constants = {}
    #         try:
    #             constants = json.loads(self.get_argument('constants'))
    #         except:
    #             pass

    #         # Experimental: python mqtt client 不支持同一topic添加多个callback，尝试实现
    #         def single_callback(client, userdata, msg):

    #             # 0.1.19
    #             payload = json.loads(msg.payload)
    #             if "content" in payload:
    #                 message = payload["content"]
    #             else:
    #                 message = payload
    #             message.update(constants)

    #             # message = json.loads(msg.payload)["content"]
    #             # message.update(constants)

    #             result = self.service(srv_name, message)
    #             if result is not None:
    #                 # Decode input
    #                 self.mqtt.publish(pub_topic, json.dumps(
    #                     {'content': result}, ensure_ascii=False), qos)

    #         cbs = []
    #         if sub_topic in self.srv.gcbs:
    #             cbs = self.srv.gcbs[sub_topic] + [single_callback]
    #             print("Add callback at " + sub_topic)
    #         else:
    #             cbs = [single_callback]
    #         self.srv.gcbs[sub_topic] = cbs

    #         def on_subs_callback(client, userdata, msg):
    #             for cb in cbs:
    #                 cb(client, userdata, msg)

    #         self.mqtt.message_callback_add(sub_topic, on_subs_callback)
    #         # Sub input
    #         print("MQTT Service Subscribe: {0}, {1}, {2}, {3}, Added service: {4}".format(
    #             self._dev_id, op, sub_topic, pub_topic, srv_name))

    #         # Save
    #         flow_models_JSON = self.sql.query(
    #             Flow).filter_by(id=self.model['id']).first()
    #         flow_model = {"_dev_id": self.model['id'], "srv_name": srv_name,
    #                       "sub_topic": sub_topic, "pub_topic": pub_topic, "qos": qos}

    #         if flow_models_JSON is None:
    #             flow_models = [flow_model]
    #         else:
    #             flow_models = json.loads(flow_models_JSON.content)
    #             if flow_model in flow_models:
    #                 return
    #             else:
    #                 flow_models.append(flow_model)

    #         if flow_models_JSON is None:
    #             self.sql.add(
    #                 Flow(id=self.model['id'], content=json.dumps(flow_models)))
    #         else:
    #             flow_models_JSON.content = json.dumps(flow_models)
    #         self.sql.commit()
    #         print("Flow Added")

    #     elif (op == "remove"):
    #         print("MQTT Service Unsubscribe")
    #         flow_models_JSON = self.sql.query(
    #             Flow).filter_by(id=self.model['id']).first()
    #         if flow_models_JSON is not None:
    #             flow_models = json.loads(flow_models_JSON.content)
    #             for model in flow_models:
    #                 self.mqtt.unsubscribe(model['sub_topic'])
    #                 self.mqtt.message_callback_remove(model['sub_topic'])
    #                 # 0.1.19
    #                 if model['sub_topic'] in self.srv.gcbs:
    #                     print("Remove global callback array at " +
    #                           model['sub_topic'])
    #                     del self.srv.gcbs[model['sub_topic']]
    #                 flow_models.remove(model)
    #                 print("Flow Deleted: " + model['sub_topic'])
    #             flow_models_JSON.content = json.dumps(flow_models)
    #         self.sql.commit()

    # def deleteAllFlows(self):
    #     flows = self.sql.query(Flow).all()
    #     try:
    #         if len(flows) > 0:
    #             for fs in flows:
    #                 flowModels = json.loads(fs.content)
    #                 for f in flowModels:
    #                     self.mqtt.unsubscribe(f['sub_topic'])
    #                     self.mqtt.message_callback_remove(f['sub_topic'])
    #                     # 0.1.19
    #                     if f['sub_topic'] in self.srv.gcbs:
    #                         print("Remove global callback array at " +
    #                               f['sub_topic'])
    #                         del self.srv.gcbs[f['sub_topic']]
    #                 self.sql.delete(fs)
    #         self.sql.commit()
    #     except Exception as e:
    #         return "Failed to delete"
    #     return "OK"

    # def getAllFlows(self):
    #     try:
    #         res = self.sql.query(Flow).all()[0].content
    #         return json.dumps(res)
    #     except Exception as e:
    #         return []


class Ping(CustomHandler):
    def get(self):
        self.write("OK")

# # 读取/操作

# # GET: {srv-name}/property/{property-name}
# # POST: {srv-name}/action/{function-name}
# # POST & MQTT: {srv-name}/event/{event-name}

# 【弃用】
# class DevProperty(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv

#     def post(self, key: str):
#         # self.write(self.dev.ReadProperty(key, content))
#         self.write(json.dumps({key: self.dev.resources[key]}))


# class DevService(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv

#     def get(self, srv_name: str):
#         self.write(json.dumps(self.srv.selector(srv_name)))

#     def post(self, srv_name: str):
#         args = self.get_argument_dict()
#         self.write(json.dumps(self.dev.service(srv_name, args)))


class DevSelector(CustomHandler):
    def initialize(self, dev, srv):
        self.dev = dev
        self.srv = srv
        self.mqtt = srv.mqtt

    def post(self, func_name: str):
        args = self.get_argument_dict()
        self.write(json.dumps(self.dev.selector(func_name, args)))


# class DevEvent(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv
#         self.mqtt = srv.mqtt

#     def post(self, event_name: str):
#         setup = int(self.get_argument('setup'))  # 0 = off, 1 = on
#         pub_topic = self.get_argument('pub_topic')
#         qos = 1
#         try:
#             qos = int(self.get_argument('qos'))
#         except:
#             pass

#         print("Event Subscribed: {0}, {1}".format(setup, pub_topic))

#         def on_event_callback(content):
#             if (setup > 0):
#                 print("Event auto-pubed: {0}, {1}".format(setup, pub_topic))
#                 self.mqtt.publish(pub_topic, json.dumps(
#                     {'content': content}, ensure_ascii=False), qos)

#         if not hasattr(self.srv.app, 'ecbs'):
#             print("ecb not initiated")
#             self.write("Error: ecb not initiated")
#         else:
#             self.srv.app.ecbs[event_name] = on_event_callback
#             self.write("OK")

# # MQTT 操作 (Flow)


# class DevMqttOp(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv
#         self.mqtt = srv.mqtt

#     def post(self, srv_name: str):
#         self.dev.flow(srv_name=srv_name, **json_decode(self.request.body))
#         self.write("OK")

#     # delete all flows
#     def delete(self, op: str):
#         if op == "all":
#             self.write(self.dev.deleteAllFlows())

#     # get all flows
#     def get(self, op: str):
#         if op == "all":
#             res = self.dev.getAllFlows()
#             if len(res) == 0:
#                 res = '[]'

#             self.write(res)

# # Protocol 协议服务专用
# class DevProc(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv

#     # Add new model
#     def post(self):
#         print(self.request.headers['Content-Type'])
#         model = self.sql.query(Entry).filter_by(id=self.get_argument('id')).first()

#         if model == None:
#             # Add
#             model = Entry(id=self.get_argument('id'), content=self.request.body.decode("utf-8"))
#             print("Device Added")
#             self.sql.add(model)
#         else:
#             # Update
#             model.content = self.request.body.decode("utf-8")
#             print("Device exists, update")
#         self.sql.commit()
#         self.set_status(200)
#         self.write('{0}'.format(model))

#     # Get all models
#     def get(self):
#         print("Get Devices")
#         model = self.sql.query(Entry).all()
#         # stop()
#         # model = [c.content for c in self.sql.query(Entry).all()]
#         self.set_status(200)
#         self.write("{0}".format(model))

# class DevProcID(CustomHandler):
#     def initialize(self, dev, srv):
#         self.dev = dev
#         self.srv = srv

#     # Delete model
#     def delete(self, id):
#         print("Delete Device - ID: "+ id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         self.sql.delete(model)
#         self.sql.commit()
#         self.set_status(200)
#         self.write("{0}".format(model))

#     # Update model
#     def post(self, id):
#         print("Update Device - ID: "+ id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         model.content = self.request.body.decode("utf-8")
#         self.sql.commit()
#         self.set_status(200)
#         self.write("{0}".format(model))

#     # Get model
#     def get(self, id: str):
#         print("Get Entry - ID: "+ id)
#         model = self.sql.query(Entry).filter_by(id=id).first()
#         self.set_status(200)
#         self.write("{0}".format(model))
